@@ -18,7 +18,7 @@
  *
  *
  * 
- * $Id: tkText.h,v 1.24 1995/09/16 00:25:01 ilya Exp $
+ * $Id: tkText.h,v 1.23 1995/09/09 06:37:48 ilya Exp ilya $
  *
  * 
  */
@@ -555,6 +555,11 @@ typedef struct TkText {
     TkTextIndex selIndex;	/* Used during multi-pass selection retrievals.
 				 * This index identifies the next character
 				 * to be returned from the selection. */
+    int selRest;		/* Used during multi-pass selection retrievals.
+				 * The number of characters in string
+				 * representation of the next text
+				 * position after selIndex that are
+				 * already written. */
     int abortSelections;	/* Set to 1 whenever the text is modified
 				 * in a way that interferes with selection
 				 * retrieval:  used to abort incremental
@@ -615,6 +620,8 @@ typedef struct TkText {
 				 * horizontal scrollbar when view changes. */
     LangCallback *yScrollCmd;	/* Prefix of command to issue to update
 				 * vertical scrollbar when view changes. */
+    int getFlags;		/* Flags for text extraction;  see below for
+				 * definitions. */
     int flags;			/* Miscellaneous flags;  see below for
 				 * definitions. */
 } TkText;
@@ -642,6 +649,14 @@ typedef struct TkText {
 #define UPDATE_SCROLLBARS	0x10
 #define NEED_REPICK		0x20
 
+#define TKT_GET_STRICT		1
+#define TKT_GET_VERBOSE		2
+
+#define TKT_GET_NOCASE		0x8000
+#define TKT_GET_OFFSET		0x4000
+#define TKT_GET_BUFFER		0x2000
+#define TKT_GET_DSTRING		0x1000
+
 /*
  * Records of the following type define segment types in terms of
  * a collection of procedures that may be called to manipulate
@@ -916,12 +931,12 @@ extern int		TkBTreeBlockStartDeleteProc _ANSI_ARGS_((
 extern void		TkTextIndexToExternalIndex _ANSI_ARGS_((
 			    TkTextBTree tree, TkTextIndex *indexPtr, 
 			    TkTextIndex *resIndexPtr));
-void			TkIBTreeTextAppend _ANSI_ARGS_((
+int			TkIBTreeTextAppend _ANSI_ARGS_((
 			    TkTextSegment *segPtr, int start, int end,
-			    Tcl_DString *dstringPtr));
-void			TkBTreeLineTextAppend _ANSI_ARGS_((
+			    void *targetPtr, int flags, int* offsetPtr));
+int			TkBTreeLineTextAppend _ANSI_ARGS_((
 			    TkTextLine *linePtr, int start, int end,
-			    Tcl_DString *dstringPtr));
+			    void *targetPtr, int flags, int* offsetPtr));
 void			TkIBTreeFixTags _ANSI_ARGS_((
 			    TkTextSegment *segPtr, TkTextTag **tagPtrPtr,
 		            int *nPtr));
@@ -943,6 +958,7 @@ struct DLine *		TkTextGetDLine _ANSI_ARGS_((TkText *textPtr,
 int			TkIBTreeTrim _ANSI_ARGS_((TkText *textPtr,
 			    TkTextSegment *segPtr));
 
+
 #ifdef LangIsPerl
 #include "tkPort.h"
 #include "tkVMacro.h"
