@@ -23,7 +23,7 @@ static char sccsid[] = "@(#) tkText.c 1.80 95/07/20 09:42:42";
 #include "tkInt.h"
 #include "tkText.h"
 
-static char rcsid[] = "$Id: tkText.c,v 1.20 1995/09/16 00:23:57 ilya Exp $";
+static char rcsid[] = "$Id: tkText.c,v 1.19 1995/09/09 06:44:23 ilya Exp ilya $";
 
 /*
  * Information used to parse text configuration options:
@@ -51,6 +51,8 @@ static Tk_ConfigSpec configSpecs[] = {
 	DEF_TEXT_FONT, Tk_Offset(TkText, fontPtr), 0},
     {TK_CONFIG_COLOR, "-foreground", "foreground", "Foreground",
 	DEF_TEXT_FG, Tk_Offset(TkText, fgColor), 0},
+    {TK_CONFIG_INT, "-getflags", "getflags", "getFlags",
+	0, Tk_Offset(TkText, getFlags), 0},
     {TK_CONFIG_PIXELS, "-height", "height", "Height",
 	DEF_TEXT_HEIGHT, Tk_Offset(TkText, height), 0},
     {TK_CONFIG_COLOR, "-highlightbackground", "highlightBackground",
@@ -293,6 +295,7 @@ Tk_TextCmd(clientData, interp, argc, argv)
     textPtr->takeFocus = NULL;
     textPtr->xScrollCmd = NULL;
     textPtr->yScrollCmd = NULL;
+    textPtr->getFlags = 0;
     textPtr->flags = 0;
 
     /*
@@ -503,6 +506,9 @@ TextWidgetCmd(clientData, interp, argc, argv)
 	}
     } else if ((c == 'g') && (strncmp(argv[1], "get", length) == 0)) {
         Tcl_DString dstring;
+	TkTextLine *linePtr;
+	int offset, dummy;
+
 	if ((argc != 3) && (argc != 4)) {
 	    Tcl_AppendResult(interp, "wrong # args: should be \"",
 		    argv[0], " get index1 ?index2?\"", (char *) NULL);
@@ -524,32 +530,19 @@ TextWidgetCmd(clientData, interp, argc, argv)
 	if (TkTextIndexCmp(&index1, &index2) >= 0) {
 	    goto done;
 	}
+	linePtr = index1.linePtr;
+	offset = index1.charIndex; 
 	Tcl_DStringInit(&dstring);
-	while (1) {
-	    int offset, last, savedChar;
-	    TkTextSegment *segPtr;
-
-	    segPtr = TkTextIndexToSeg(&index1, &offset);
-	    last = segPtr->size;
-	    if (index1.linePtr == index2.linePtr) {
-		int last2;
-
-		if (index2.charIndex == index1.charIndex) {
-		    break;
-		}
-		last2 = index2.charIndex - index1.charIndex + offset;
-		if (last2 < last) {
-		    last = last2;
-		}
-	    }
-	    if (segPtr->typePtr == &tkTextCharType) {
-		Tcl_DStringAppend(&dstring, segPtr->body.chars + offset,
-				  last - offset);
-	    } else if (segPtr->typePtr == &tkTextIBTreeType) {
-	      TkIBTreeTextAppend(segPtr, offset, last, &dstring);
-	    }
-	    TkTextIndexForwChars(&index1, last-offset, &index1);
+	while (linePtr != index2.linePtr) {
+	  TkBTreeLineTextAppend(linePtr, offset, -1, 
+				(void*)&dstring, 
+				textPtr->getFlags | TKT_GET_DSTRING, NULL);
+	  offset = 0;
+	  linePtr = TkBTreeNextLine(linePtr);
 	}
+	TkBTreeLineTextAppend(linePtr, offset, index2.charIndex,
+			      (void*)&dstring, 
+			      textPtr->getFlags | TKT_GET_DSTRING, NULL);
 	Tcl_AppendResult(interp, dstring.string, (char *) NULL);
 	Tcl_DStringFree(&dstring);
     } else if ((c == 'i') && (strncmp(argv[1], "index", length) == 0)
@@ -1286,9 +1279,10 @@ TextFetchSelection(clientData, offset, buffer, maxBytes)
 {
     register TkText *textPtr = (TkText *) clientData;
     TkTextIndex eof;
-    int count, chunkSize, offsetInSeg;
+    int count, chunkSize, offsetInSeg, startOff, endOff;
     TkTextSearch search;
     TkTextSegment *segPtr;
+    TkTextLine *linePtr;
 
     if (!textPtr->exportSelection) {
 	return -1;
@@ -1305,6 +1299,7 @@ TextFetchSelection(clientData, offset, buffer, maxBytes)
     if (offset == 0) {
 	TkTextMakeIndex(textPtr->tree, 0, 0, &textPtr->selIndex);
 	textPtr->abortSelections = 0;
+	textPtr->selRest = 0;
     } else if (textPtr->abortSelections) {
 	return 0;
     } else {
@@ -1321,7 +1316,7 @@ TextFetchSelection(clientData, offset, buffer, maxBytes)
 		return 0;
 	    }
 	}
-	textPtr->selIndex = search.curIndex;
+	textPtr->selIndex = search.extIndex;
     }
 
     /*
@@ -1331,6 +1326,34 @@ TextFetchSelection(clientData, offset, buffer, maxBytes)
      */
 
     count = 0;
+    if (textPtr->selRest) {	/* Need to skip first several characters. */
+        Tcl_DString dString;
+	int i;
+	char *s;
+
+	Tcl_DStringInit(&dString);
+	TkBTreeLineTextAppend(linePtr, textPtr->selIndex.charIndex,
+			      textPtr->selIndex.charIndex + 1, 
+			      (void*)&dString, 
+			      textPtr->getFlags | TKT_GET_DSTRING,
+			      NULL);
+	s = Tcl_DStringValue(&dString) + textPtr->selRest;
+	i = Tcl_DStringLength(&dString) - textPtr->selRest;
+	if (i > maxBytes) {
+	  textPtr->selRest += maxBytes;
+	  i = maxBytes;
+	} else {
+	  textPtr->selRest = 0;
+	  textPtr->selIndex.charIndex++;
+	}
+	maxBytes -= i;
+	count += i;
+	while (i--) {
+	  *buffer++ = *s++;
+	}
+	textPtr->selRest = maxBytes;
+	Tcl_DStringFree(&dString);
+    }
     while (1) {
 	/*
 	 * Find the end of the current range of selected text.
@@ -1341,41 +1364,62 @@ TextFetchSelection(clientData, offset, buffer, maxBytes)
 	}
 
 	/*
-	 * Copy information from character segments into the buffer
+	 * Copy information from character lines into the buffer
 	 * until either we run out of space in the buffer or we get
 	 * to the end of this range of text.
 	 */
 
+	linePtr =  textPtr->selIndex.linePtr;
+	startOff = textPtr->selIndex.charIndex;
+	endOff = -1;
 	while (1) {
+	    int pos, m = maxBytes;
+
 	    if (maxBytes == 0) {
 		goto done;
 	    }
-	    segPtr = TkTextIndexToSeg(&textPtr->selIndex, &offsetInSeg);
-	    chunkSize = segPtr->size - offsetInSeg;
-	    if (chunkSize > maxBytes) {
-		chunkSize = maxBytes;
+	    if (linePtr == search.extIndex.linePtr) {
+	      endOff = search.extIndex.charIndex;
 	    }
-	    if (textPtr->selIndex.linePtr == search.extIndex.linePtr) {
-		int leftInRange;
+	    pos = TkBTreeLineTextAppend(linePtr, startOff, endOff,
+					(void *)buffer, 
+					textPtr->getFlags | TKT_GET_BUFFER,
+					&maxBytes);
+	    buffer += m - maxBytes;
+	    count  += m - maxBytes;
+	    /* Need to check the status: could reach the end of line,
+	     * or the end of buffer (some place may been left in buffer
+	     * if the representation of the last character takes more
+	     * than 1 byte). */
+	    if (pos == endOff) {	/* Done with this range */
+	      break;
+	    } else if (pos == TkBTreeCharsInLine(linePtr)) { /* End of line */
+	      linePtr = TkBTreeNextLine(linePtr);
+	      startOff = 0;
+	      continue;
+	    } else {		/* Did not reach the end of line */
+	      if (maxBytes) {	/* Next index takes several bytes in
+				 * string representation. */
+		Tcl_DString dString;
+		int i = maxBytes;
+		char *s;
 
-		leftInRange = search.extIndex.charIndex
-			- textPtr->selIndex.charIndex;
-		if (leftInRange < chunkSize) {
-		    chunkSize = leftInRange;
-		    if (chunkSize <= 0) {
-			break;
-		    }
+		Tcl_DStringInit(&dString);
+		TkBTreeLineTextAppend(linePtr, pos, pos + 1, 
+				      (void*)&dString, 
+				      textPtr->getFlags | TKT_GET_DSTRING,
+				      NULL);
+		s = Tcl_DStringValue(&dString);
+		while (i--) {
+		  *buffer++ = *s++;
 		}
+		textPtr->selRest = maxBytes;
+		Tcl_DStringFree(&dString);
+	      }
+	      textPtr->selIndex.linePtr = linePtr;
+	      textPtr->selIndex.charIndex = pos;
+	      goto done;
 	    }
-	    if (segPtr->typePtr == &tkTextCharType) {
-		memcpy((VOID *) buffer, (VOID *) (segPtr->body.chars
-			+ offsetInSeg), (size_t) chunkSize);
-		buffer += chunkSize;
-		maxBytes -= chunkSize;
-		count += chunkSize;
-	    }
-	    TkTextIndexForwChars(&textPtr->selIndex, chunkSize,
-		    &textPtr->selIndex);
 	}
 
 	/*
@@ -1385,7 +1429,7 @@ TextFetchSelection(clientData, offset, buffer, maxBytes)
 	if (!TkBTreeNextTag(&search)) {
 	    break;
 	}
-	textPtr->selIndex = search.curIndex;
+	textPtr->selIndex = search.extIndex;
     }
 
     done:
@@ -1511,7 +1555,7 @@ TextSearchCmd(textPtr, interp, argc, argv)
     Tcl_DString line, patDString;
     TkTextSegment *segPtr;
     TkTextLine *linePtr;
-    Tcl_RegExp regexp1 = NULL;		/* Initialization needed only to
+    Tcl_RegExp regExp = NULL;		/* Initialization needed only to
 					 * prevent compiler warning. */
 
     /*
@@ -1550,7 +1594,7 @@ TextSearchCmd(textPtr, interp, argc, argv)
 	} else if ((c == 'f') && (strncmp(argv[i], "-forwards", length) == 0)) {
 	    backwards = 0;
 	} else if ((c == 'n') && (strncmp(argv[i], "-nocase", length) == 0)) {
-	    noCase = 1;
+	    noCase = TKT_GET_NOCASE;
 	} else if ((c == 'r') && (strncmp(argv[i], "-regexp", length) == 0)) {
 	    exact = 0;
 	} else if ((c == '-') && (strncmp(argv[i], "--", length) == 0)) {
@@ -1618,8 +1662,8 @@ TextSearchCmd(textPtr, interp, argc, argv)
     if (exact) {
 	patLength = strlen(pattern);
     } else {
-	regexp1 = Tcl_RegExpCompile(interp, pattern);
-	if (regexp1 == NULL) {
+	regExp = Tcl_RegExpCompile(interp, pattern);
+	if (regExp == NULL) {
 	    return TCL_ERROR;
 	}
     }
@@ -1642,30 +1686,40 @@ TextSearchCmd(textPtr, interp, argc, argv)
 	 */
 
 	linePtr = TkBTreeFindLine(textPtr->tree, lineNum);
-	for (segPtr = linePtr->segPtr; segPtr != NULL;
-		segPtr = segPtr->nextPtr) {
-	    if (segPtr->typePtr != &tkTextCharType) {
-		continue;
+
+	/*
+	 * The starting line is tricky: the first time we see it
+	 * we check one part of the line, and the second pass through
+	 * we check the other part of the line.  
+	 */
+
+	if (lineNum == startingLine) {
+	    passes++;
+	    TkBTreeLineTextAppend(linePtr, 0, startingChar, (void*)&line, 
+				  textPtr->getFlags | noCase | TKT_GET_DSTRING,
+				  NULL);
+	    firstChar = Tcl_DStringLength(&line);
+	    TkBTreeLineTextAppend(linePtr, startingChar, -1, (void*)&line, 
+				  textPtr->getFlags | noCase | TKT_GET_DSTRING,
+				  NULL);
+	    if ((passes == 1) ^ backwards) {
+		lastChar = INT_MAX;
+	    } else {
+	      lastChar = firstChar;
+	      firstChar = 0;
 	    }
-	    Tcl_DStringAppend(&line, segPtr->body.chars, segPtr->size);
+	} else {
+	    TkBTreeLineTextAppend(linePtr, 0, -1, (void*)&line, 
+				  textPtr->getFlags | noCase | TKT_GET_DSTRING,
+				  NULL);
+	    firstChar = 0;
+	    lastChar = INT_MAX;
 	}
 	if (!exact) {
 	    Tcl_DStringSetLength(&line, Tcl_DStringLength(&line)-1);
 	}
 	startOfLine = Tcl_DStringValue(&line);
 
-	/*
-	 * If we're ignoring case, convert the line to lower case.
-	 */
-
-	if (noCase) {
-	    for (p = Tcl_DStringValue(&line); *p != 0; p++) {
-		if (isupper(UCHAR(*p))) {
-		    *p = tolower(UCHAR(*p));
-		}
-	    }
-	}
-
 	/*
 	 * Check for matches within the current line.  If so, and if we're
 	 * searching backwards, repeat the search to find the last match
@@ -1673,48 +1727,6 @@ TextSearchCmd(textPtr, interp, argc, argv)
 	 */
 
 	matchChar = -1;
-	firstChar = 0;
-	lastChar = INT_MAX;
-	if (lineNum == startingLine) {
-	    int indexInDString;
-
-	    /*
-	     * The starting line is tricky: the first time we see it
-	     * we check one part of the line, and the second pass through
-	     * we check the other part of the line.  We have to be very
-	     * careful here because there could be embedded windows or
-	     * other things that are not in the extracted line.  Rescan
-	     * the original line to compute the index in it of the first
-	     * character.
-	     */
-
-	    indexInDString = startingChar;
-	    for (segPtr = linePtr->segPtr, leftToScan = startingChar;
-		    leftToScan > 0; segPtr = segPtr->nextPtr) {
-		if (segPtr->typePtr != &tkTextCharType) {
-		    indexInDString -= segPtr->size;
-		}
-		leftToScan -= segPtr->size;
-	    }
-
-	    passes++;
-	    if ((passes == 1) ^ backwards) {
-		/*
-		 * Only use the last part of the line.
-		 */
-
-		firstChar = indexInDString;
-		if (firstChar >= Tcl_DStringLength(&line)) {
-		    goto nextLine;
-		}
-	    } else {
-		/*
-		 * Use only the first part of the line.
-		 */
-
-		lastChar = indexInDString;
-	    }
-	}
 	do {
 	    int thisLength;
 	    if (exact) {
@@ -1728,7 +1740,7 @@ TextSearchCmd(textPtr, interp, argc, argv)
 		char *start, *end;
 		int match;
 
-		match = Tcl_RegExpExec(interp, regexp1,
+		match = Tcl_RegExpExec(interp, regExp,
 			startOfLine + firstChar, startOfLine);
 		if (match < 0) {
 		    code = TCL_ERROR;
@@ -1737,7 +1749,7 @@ TextSearchCmd(textPtr, interp, argc, argv)
 		if (!match) {
 		    break;
 		}
-		Tcl_RegExpRange(regexp1, 0, &start, &end);
+		Tcl_RegExpRange(regExp, 0, &start, &end);
 		i = start - startOfLine;
 		thisLength = end - start;
 	    }
@@ -1758,29 +1770,23 @@ TextSearchCmd(textPtr, interp, argc, argv)
 	if (matchChar >= 0) {
 	    /*
 	     * The index information returned by the regular expression
-	     * parser only considers textual information:  it doesn't
-	     * account for embedded windows or any other non-textual info.
-	     * Scan through the line's segments again to adjust both
-	     * matchChar and matchCount.
+	     * parser considers string representation. Now find the
+	     * position in the text widget.
 	     */
 
-	    for (segPtr = linePtr->segPtr, leftToScan = matchChar;
-		    leftToScan >= 0; segPtr = segPtr->nextPtr) {
-		if (segPtr->typePtr != &tkTextCharType) {
-		    matchChar += segPtr->size;
-		    continue;
-		}
-		leftToScan -= segPtr->size;
-	    }
-	    for (leftToScan += matchLength; leftToScan > 0;
-		    segPtr = segPtr->nextPtr) {
-		if (segPtr->typePtr != &tkTextCharType) {
-		    matchLength += segPtr->size;
-		    continue;
-		}
-		leftToScan -= segPtr->size;
-	    }
-	    TkTextMakeIndex(textPtr->tree, lineNum, matchChar, &index);
+	    int offset = matchChar;
+	    int posStart = TkBTreeLineTextAppend(linePtr, 0, -1, NULL,
+						 textPtr->getFlags | TKT_GET_OFFSET,
+						 &offset);
+
+	    offset += matchLength;
+	    matchLength = TkBTreeLineTextAppend(linePtr, posStart, -1, NULL,
+						textPtr->getFlags | TKT_GET_OFFSET,
+						&offset) 
+	      - posStart;
+	    if (offset) matchLength++; /* Include partially matched
+					* character. */
+	    TkTextMakeIndex(textPtr->tree, lineNum, posStart, &index);
 	    if (!searchWholeText) {
 		if (!backwards && (TkTextIndexCmp(&index, &stopIndex) >= 0)) {
 		    goto done;
@@ -1841,8 +1847,8 @@ TextSearchCmd(textPtr, interp, argc, argv)
     if (noCase) {
 	Tcl_DStringFree(&patDString);
     }
-    if (regexp1) {
-	Lang_FreeRegExp(regexp1);
+    if (regExp) {
+	Lang_FreeRegExp(regExp);
     }
     return code;
 }
